AUTH_FILE = "sorcerun_auth.json"
MONGOD_HOST = "127.0.0.1"
MONGOD_PORT = 27017
GRID_OUTPUTS = "grid_outputs"
RUNS_DIR = "runs"
FILE_STORAGE_ROOT = "file_storage"
TEMP_CONFIGS_DIR = "temp_configs"
TEMPLATE_FILES = [
    "main/main.py",
    "main/config.py",
    "main/grid_config.py",
]
TIME_FORMAT = "%Y-%m-%d-%H-%M-%S"
